/*
 * imgRGB.c --
 *
 *	A photo image file handler for IRIS RGB files.
 *
 * Written by:
 *	Jan Nijtmans
 *	NICI (Nijmegen Institute of Cognition and Information)
 *	email: nijtmans@nici.kun.nl
 *	url:   http://www.cogsci.kun.nl/~nijtmans/
 *
 */

#include "tcl.h"
#include "tk.h"
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

/* constants used only in this file */

#define MAX_BUFFER 4096

/*
 * This structure is needed to access strings in the same 
 * way as a file
 */
typedef struct RGB_DString {
  char *data;  /* the data string          */
  char *p;     /* character at current pos */
} RGB_DString;

/*
 * The format record for the RGB file format:
 */

static int		FileMatchRGB _ANSI_ARGS_((FILE *f, char *fileName,
			    char *formatString, int *widthPtr,
			    int *heightPtr));
static int      	StringMatchRGB _ANSI_ARGS_(( char *string,
		            char *formatString, int *widthPtr, int *heightPtr));
static int	        StringReadRGB _ANSI_ARGS_((Tcl_Interp *interp,
			    char *string, char *formatString,
              		    Tk_PhotoHandle imageHandle, int destX, int destY,
		            int width, int height, int srcX, int srcY));
static int		FileReadRGB  _ANSI_ARGS_((Tcl_Interp *interp,
			    FILE *f, char *fileName, char *formatString,
			    Tk_PhotoHandle imageHandle, int destX, int destY,
			    int width, int height, int srcX, int srcY));
static int		ReadRGB _ANSI_ARGS_((Tcl_Interp *interp, FILE *f,
			    char *fileName, RGB_DString *dataPtr,
			    char *formatString, Tk_PhotoHandle imageHandle,
			    int destX, int destY, int width, int height,
			    int srcX, int srcY));
static int	        StringWriteRGB _ANSI_ARGS_((Tcl_Interp *interp,
               		    Tcl_DString *dataPtr, char *formatString,
		            Tk_PhotoImageBlock *blockPtr));
static int              FileWriteRGB _ANSI_ARGS_((Tcl_Interp *interp,
                            char *fileName, char *formatString,
                            Tk_PhotoImageBlock *blockPtr));
static int		WriteRGB _ANSI_ARGS_((Tcl_Interp *interp,
			    char *fileName, Tcl_DString *dataPtr, char *formatString,
			    Tk_PhotoImageBlock *blockPtr));

Tk_PhotoImageFormat imgFmtRGB = {
    "RGB",			/* name */
    FileMatchRGB,		/* fileMatchProc */
    StringMatchRGB,		/* stringMatchProc */
    FileReadRGB,		/* fileReadProc */
    StringReadRGB,		/* stringReadProc */
    FileWriteRGB,		/* fileWriteProc */
    StringWriteRGB,		/* stringWriteProc */
};

/*
 * Prototypes for local procedures defined in this file:
 */

static int	ReadRGBFileHeader _ANSI_ARGS_((FILE *f, 
			RGB_DString *dataPtr, int *widthPtr,
			int *heightPtr));
static char *	GetType _ANSI_ARGS_((char *colorDefn, int *type_ret));
static char *	GetColor _ANSI_ARGS_((char *colorDefn, char *colorName, int *type_ret));
static char *	Gets _ANSI_ARGS_((char *buffer, int size, FILE *f, RGB_DString *dataPtr));

/*
 *----------------------------------------------------------------------
 * Gets
 *
 *      Allows other routines to get their data from files as well
 *      as Tcl_DStrings. If FILE *f == NULL, then data is fetched
 *      from dataPtr, otherwise data is read from FILE *f via fgets().
 *
 * Results:
 *      same as fgets: NULL pointer on any error, otherwise
 *      it returns buffer
 *
 * Side effects:
 *      The access position of the file changes OR
 *      The access pointer dataPtr->p is changed
 *
 *----------------------------------------------------------------------
 */

static char *
Gets(buffer, size, f, dataPtr) 
    char *buffer;
    int size;
    FILE *f;
    struct RGB_DString *dataPtr;
{
    char *p;

    if (f != (FILE *)NULL) {
	/* read data from file */
	return fgets(buffer, size, f);
    } else {
	/* emulate premature EOF */
	if (*dataPtr->p == 0) {
	    return (char *)NULL;
	}

	p = buffer;
	size--;
	while(size--) {
	    if (*dataPtr->p == 0) {
		*p = 0;
		return buffer;
	    } else {
		if (*dataPtr->p == '\n') {
		    *p++ = *dataPtr->p++;
		    *p = 0;
		    return buffer;
		}
	    }
	    *p++ = *dataPtr->p++;
	}
	*p = 0;
	return buffer;
    }
}


/*
 *----------------------------------------------------------------------
 *
 * StringMatchRGB --
 *
 *	This procedure is invoked by the photo image type to see if
 *	a datastring contains image data in RGB format.
 *
 * Results:
 *	The return value is >0 if the first characters in data look
 *	like RGB data, and 0 otherwise.
 *
 * Side effects:
 *	none
 *
 *----------------------------------------------------------------------
 */
static int
StringMatchRGB(string, formatString, widthPtr, heightPtr)
    char *string;               /* The data supplied by the image */
    char *formatString;		/* User-specified format string, or NULL. */
    int *widthPtr, *heightPtr;	/* The dimensions of the image are
				 * returned here if the file is a valid
				 * raw RGB file. */
{
    struct RGB_DString d;
    int numColors, byteSize;

    d.data = string;
    d.p = string;
    return ReadRGBFileHeader(NULL, &d, widthPtr, heightPtr);
}


/*
 *----------------------------------------------------------------------
 *
 * FileMatchRGB --
 *
 *	This procedure is invoked by the photo image type to see if
 *	a file contains image data in RGB format.
 *
 * Results:
 *	The return value is >0 if the first characters in file "f" look
 *	like RGB data, and 0 otherwise.
 *
 * Side effects:
 *	The access position in f may change.
 *
 *----------------------------------------------------------------------
 */

static int
FileMatchRGB(f, fileName, formatString, widthPtr, heightPtr)
    FILE *f;			/* The image file, open for reading. */
    char *fileName;		/* The name of the image file. */
    char *formatString;		/* User-specified format string, or NULL. */
    int *widthPtr, *heightPtr;	/* The dimensions of the image are
				 * returned here if the file is a valid
				 * raw RGB file. */
{
    int numColors, byteSize;
    return ReadRGBFileHeader(f, NULL, widthPtr, heightPtr);
}


/*
 *----------------------------------------------------------------------
 *
 * ReadRGB --
 *
 *	This procedure is called by the photo image type to read
 *	RGB format data from a file or string and write it into a
 *	given photo image.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	The access position in file f is changed (if read from file)
 *	and new data is added to the image given by imageHandle.
 *
 *----------------------------------------------------------------------
 */

typedef struct myblock {
    Tk_PhotoImageBlock pub;
    int dummy; /* extra space for offset[3], if not included already
		  in Tk_PhotoImageBlock */
} myblock;

static int
ReadRGB(interp, f, fileName, dataPtr, formatString, imageHandle, destX, destY,
	width, height, srcX, srcY)
    Tcl_Interp *interp;		/* Interpreter to use for reporting errors. */
    FILE *f;			/* The image file, open for reading. */
    char *fileName;		/* The name of the image file. */
    struct RGB_DString *dataPtr;
    char *formatString;		/* User-specified format string, or NULL. */
    Tk_PhotoHandle imageHandle;	/* The photo image to write into. */
    int destX, destY;		/* Coordinates of top-left pixel in
				 * photo image to be written to. */
    int width, height;		/* Dimensions of block of photo image to
				 * be written to. */
    int srcX, srcY;		/* Coordinates of top-left pixel to be used
				 * in image being read. */
{
    int fileWidth = 0, fileHeight = 0, numColors = 0, byteSize = 0;
    int h, type;
    unsigned char *pixelPtr;
    myblock block;
    Tcl_HashTable colorTable;
    Tk_Window tkwin = Tk_MainWindow(interp);
    Display *display = Tk_Display(tkwin);
    Colormap colormap = Tk_Colormap(tkwin);
    int depth = Tk_Depth(tkwin);
    char *p;
    char buffer[MAX_BUFFER];
    int i, isMono;
    int color1;
    unsigned int data;
    Tcl_HashEntry *hPtr;

    Tcl_InitHashTable(&colorTable, TCL_ONE_WORD_KEYS);

    switch ((Tk_Visual(tkwin))->class) {
      case StaticGray:
      case GrayScale:
	isMono = 1;
	break;
      default:
	isMono = 0;
    }

    type = ReadRGBFileHeader(f, dataPtr, &fileWidth, &fileHeight);
    if (type == 0) {
	Tcl_AppendResult(interp, "couldn't read raw RGB header from file \"",
		fileName, "\"", NULL);
	return TCL_ERROR;
    }
    if ((fileWidth <= 0) || (fileHeight <= 0)) {
	Tcl_AppendResult(interp, "RGB image file \"", fileName,
		"\" has dimension(s) <= 0", (char *) NULL);
	return TCL_ERROR;
    }
    if ((byteSize < 1) || (byteSize > 4)) {
	Tcl_AppendResult(interp, "RGB image file \"", fileName,
		"\" has invalid byte size (should be 1, 2, 3 or 4)", (char *) NULL);
	return TCL_ERROR;
    }

    if ((srcX + width) > fileWidth) {
	width = fileWidth - srcX;
    }
    if ((srcY + height) > fileHeight) {
	height = fileHeight - srcY;
    }
    if ((width <= 0) || (height <= 0)
	|| (srcX >= fileWidth) || (srcY >= fileHeight)) {
	return TCL_OK;
    }


    Tk_PhotoGetImage(imageHandle, &block.pub);
	 /* in case Tk_PhotoGetImage doesn't set this */
    block.pub.offset[3] = (block.pub.pixelSize > 3) ? 3 : 0;
    block.pub.width = width;
    block.pub.pitch = block.pub.pixelSize * fileWidth;
    block.pub.height = 1;
    block.pub.pixelPtr = (unsigned char *) ckalloc((unsigned) block.pub.pixelSize * width);

    Tk_PhotoExpand(imageHandle, destX + width, destY + height);

    ckfree((char *) block.pub.pixelPtr);
    return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * FileReadRGB --
 *
 *	This procedure is called by the photo image type to read
 *	RGB format data from a file and write it into a given
 *	photo image.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	The access position in file f is changed, and new data is
 *	added to the image given by imageHandle.
 *
 *----------------------------------------------------------------------
 */

static int
FileReadRGB(interp, f, fileName, formatString, imageHandle, destX, destY,
	width, height, srcX, srcY)
    Tcl_Interp *interp;		/* Interpreter to use for reporting errors. */
    FILE *f;			/* The image file, open for reading. */
    char *fileName;		/* The name of the image file. */
    char *formatString;		/* User-specified format string, or NULL. */
    Tk_PhotoHandle imageHandle;	/* The photo image to write into. */
    int destX, destY;		/* Coordinates of top-left pixel in
				 * photo image to be written to. */
    int width, height;		/* Dimensions of block of photo image to
				 * be written to. */
    int srcX, srcY;		/* Coordinates of top-left pixel to be used
				 * in image being read. */
{
  return ReadRGB(interp, f, fileName, NULL, formatString, imageHandle,
		 destX, destY, width, height, srcX, srcY);
}


/*
 *----------------------------------------------------------------------
 *
 * StringReadRGB --
 *
 *	This procedure is called by the photo image type to read
 *	RGB format data from a string and write it into a given
 *	photo image.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	New data is added to the image given by imageHandle.
 *
 *----------------------------------------------------------------------
 */

static int
StringReadRGB(interp, string, formatString, imageHandle, destX, destY,
	width, height, srcX, srcY)
    Tcl_Interp *interp;		/* Interpreter to use for reporting errors. */
    char *string;
    char *formatString;		/* User-specified format string, or NULL. */
    Tk_PhotoHandle imageHandle;	/* The photo image to write into. */
    int destX, destY;		/* Coordinates of top-left pixel in
				 * photo image to be written to. */
    int width, height;		/* Dimensions of block of photo image to
				 * be written to. */
    int srcX, srcY;		/* Coordinates of top-left pixel to be used
				 * in image being read. */
{
  struct RGB_DString data;

  data.data = string;
  data.p = string;
  return ReadRGB(interp, NULL, "", &data, formatString, imageHandle,
		 destX, destY, width, height, srcX, srcY);
}


/*
 *----------------------------------------------------------------------
 *
 * ReadRGBFileHeader --
 *
 *	This procedure reads the RGB header from the beginning of a
 *	RGB file and returns information from the header.
 *
 * Results:
 *	The return value is 1 if file "f" appears to start with a valid
 *      RGB header, and 0 otherwise.  If the header is valid,
 *	then *widthPtr and *heightPtr are modified to hold the
 *	dimensions of the image and *numColors holds the number of
 *	colors and byteSize the number of bytes used for 1 pixel.
 *
 * Side effects:
 *	The access position in f advances.
 *
 *----------------------------------------------------------------------
 */

#define UCHAR(c) ((unsigned char) (c))

static int
ReadRGBFileHeader(f, dataPtr, widthPtr, heightPtr)
    FILE *f;			/* Image file to read the header from */
    struct RGB_DString *dataPtr;
    int *widthPtr, *heightPtr;	/* The dimensions of the image are
				 * returned here. */
{
    char buffer[MAX_BUFFER];
    char *p;

    p = Gets(buffer,MAX_BUFFER,f,dataPtr);
    if (p == NULL) {
	return 0;
    }
    p = buffer;
    while (*p && isspace(UCHAR(*p))) {
	p++;
    }
    if (strncmp("/* RGB", p, 6) != 0) {
	return 0;
    }
    while ((p = strchr(p,'{')) == NULL) {
	p = Gets(buffer,MAX_BUFFER,f,dataPtr);
	if (p == NULL) {
	    return 0;
	}
	p = buffer;
    }
    while ((p = strchr(p,'"')) == NULL) {
	p = Gets(buffer,MAX_BUFFER,f,dataPtr);
	if (p == NULL) {
	    return 0;
	}
	p = buffer;
    }
    p++;
    while (p && *p && isspace(UCHAR(*p))) {
	p++;
    }
    *widthPtr = strtoul(p, &p, 0);
    if (p == NULL) {
	return 0;
    }
    while (p && *p && isspace(UCHAR(*p))) {
	p++;
    }
    *heightPtr = strtoul(p, &p, 0);
    if (p == NULL) {
	return 0;
    }
    while (p && *p && isspace(UCHAR(*p))) {
	p++;
    }
    return 1;
}

/*
 *----------------------------------------------------------------------
 *
 * FileWriteRGB
 *
 *	Writes a RGB image to a file. Just calls WriteRGB
 *      with appropriate arguments.
 *
 * Results:
 *	Returns the return value of WriteRGB
 *
 * Side effects:
 *	A file is (hopefully) created on success.
 *
 *----------------------------------------------------------------------
 */
static int
FileWriteRGB(interp, fileName, formatString, blockPtr)
    Tcl_Interp *interp;
    char *fileName;
    char *formatString;
    Tk_PhotoImageBlock *blockPtr;
{
    return WriteRGB(interp, fileName, (Tcl_DString *)NULL, formatString, blockPtr);
}


/*
 *----------------------------------------------------------------------
 *
 * StringWriteRGB
 *
 *	Writes a RGB image to a string. Just calls WriteRGB
 *      with appropriate arguments.
 *
 * Results:
 *	Returns the return value of WriteRGB
 *
 * Side effects:
 *	The Tcl_DString dataPtr is modified on success.
 *
 *----------------------------------------------------------------------
 */
static int	        
StringWriteRGB(interp, dataPtr, formatString, blockPtr) 
    Tcl_Interp *interp;
    Tcl_DString *dataPtr;
    char *formatString;
    Tk_PhotoImageBlock *blockPtr;
{
  return WriteRGB(interp, (char *) NULL, dataPtr, formatString, blockPtr);
}


/*
 * Yes, I know these macros are dangerous. But it should work fine
 */
#define WRITE(buf) { if (f) fputs(buf, f); else Tcl_DStringAppend(dataPtr, buf, -1);}

/*
 *----------------------------------------------------------------------
 *
 * WriteRGB
 *
 *	This procedure writes a RGB image to the file filename 
 *      (if filename != NULL) or to dataPtr.
 *
 * Results:
 *	Returns TCL_OK on success, or TCL_ERROR on error.
 *      Possible failures are:
 *      1. cannot access file (permissions or path not found)
 *      2. TkPhotoGetMask fails to retrieve the region mask
 *         for the image (should not happen)
 *
 * Side effects:
 *	varies (see StringWriteRGB and FileWriteRGB)
 *
 *----------------------------------------------------------------------
 */
static int
WriteRGB(interp, fileName, dataPtr, formatString, blockPtr)
    Tcl_Interp *interp;
    char *fileName;
    Tcl_DString *dataPtr;
    char *formatString;    
    Tk_PhotoImageBlock *blockPtr;
{
    unsigned int x, y, i, found;
    FILE *f = (FILE *) NULL;
    Tcl_HashTable colors;
    Tcl_HashEntry *entry;
    Tcl_HashSearch search;
    unsigned char *pp;
    unsigned int colvalue, ncolors, chars_per_pixel;
    char temp[9], buffer[256];

    /* open the output file (if needed) */
    if (fileName) {
      f = fopen(fileName, "w");
      if (f == (FILE *)NULL) {
	Tcl_AppendResult(interp, ": cannot open file for writing",
		(char *)NULL);
	return TCL_ERROR;
      }
    }


    /* close the file */
    if (f) {
	fclose(f);
    }
    return TCL_OK;
}
